# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

from sqlalchemy.orm import mapper, relationship, backref, clear_mappers
from sqlalchemy.orm.relationships import foreign
from sqlalchemy.sql.schema import Column, MetaData, Table, ForeignKey, \
    ForeignKeyConstraint, Index
from sqlalchemy.sql.sqltypes import String, Integer, Float, Date, Boolean

from gtfslib.model import FeedInfo, Agency, Stop, Route, Calendar, CalendarDate, \
    Trip, StopTime, Transfer, Shape, ShapePoint, Zone, FareAttribute, FareRule


# ORM Mappings
class _Orm(object):

    def __init__(self, engine, schema=None):
        self._metadata = MetaData(schema=schema)
        self.mappers = []

        _feedinfo_id_column = Column('feed_id', String, primary_key=True)
        _agency_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _route_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _feedinfo_mapper = Table('feed_info', self._metadata,
                    _feedinfo_id_column,
                    Column('feed_publisher_name', String),
                    Column('feed_publisher_url', String),
                    Column('feed_contact_email', String), # Non-standard (yet) field
                    Column('feed_contact_url', String), # Non-standard (yet) field
                    Column('feed_lang', String),
                    Column('feed_start_date', Date),
                    Column('feed_end_date', Date),
                    Column('feed_version', String))
        self.mappers.append(mapper(FeedInfo, _feedinfo_mapper, properties={
        }))

        _agency_id_column = Column('agency_id', String, primary_key=True)
        _route_agency_id_column = Column('agency_id', String, nullable=False)
        _agency_mapper = Table('agency', self._metadata,
                    _agency_feed_id_column,
                    _agency_id_column,
                    Column('agency_name', String, nullable=False),
                    Column('agency_url', String, nullable=False),
                    Column('agency_timezone', String, nullable=False),
                    Column('agency_lang', String),
                    Column('agency_phone', String),
                    Column('agency_fare_url', String),
                    Column('agency_email', String))
        self.mappers.append(mapper(Agency, _agency_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('agencies', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_agency_feed_id_column))
        }))

        _zone_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _zone_id_column = Column('zone_id', String, primary_key=True)
        _zone_mapper = Table('zones', self._metadata,
                    _zone_feed_id_column,
                    _zone_id_column)
        self.mappers.append(mapper(Zone, _zone_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('zones', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_zone_feed_id_column))
        }))

        _stop_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _stop_id_column = Column('stop_id', String, primary_key=True)
        _stop_parent_id_column = Column('parent_station_id', String, nullable=True)
        _stop_zone_id_column = Column('zone_id', String, nullable=True)
        _stop_mapper = Table('stops', self._metadata,
                    _stop_feed_id_column,
                    _stop_id_column,
                    _stop_parent_id_column,
                    Column('location_type', Integer, nullable=False),
                    Column('stop_name', String, nullable=False),
                    Column('stop_lat', Float, nullable=False),
                    Column('stop_lon', Float, nullable=False),
                    Column('wheelchair_boarding', Integer, nullable=False),
                    Column('stop_code', String),
                    Column('stop_desc', String),
                    _stop_zone_id_column,
                    Column('stop_url', String),
                    Column('stop_timezone', String),
                    ForeignKeyConstraint(['feed_id', 'parent_station_id'], ['stops.feed_id', 'stops.stop_id']),
                    ForeignKeyConstraint(['feed_id', 'zone_id'], ['zones.feed_id', 'zones.zone_id']),
                    # TODO Make those index parametrable
                    Index('idx_stops_lat', 'stop_lat'),
                    Index('idx_stops_lon', 'stop_lon'),
                    Index('idx_stops_code', 'feed_id', 'stop_code'),
                    Index('idx_stops_zone', 'feed_id', 'zone_id'),
                    Index('idx_stops_parent', 'feed_id', 'parent_station_id'))
        self.mappers.append(mapper(Stop, _stop_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('stops', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_stop_feed_id_column)),
            'sub_stops' : relationship(Stop, remote_side=[_stop_feed_id_column, _stop_parent_id_column], uselist=True,
                                       primaryjoin=(_stop_parent_id_column == foreign(_stop_id_column)) & (_stop_feed_id_column == _stop_feed_id_column)),
            'parent_station' : relationship(Stop, remote_side=[_stop_feed_id_column, _stop_id_column],
                                       primaryjoin=(_stop_id_column == foreign(_stop_parent_id_column)) & (_stop_feed_id_column == _stop_feed_id_column)),
            'zone' : relationship(Zone, backref=backref('stops', cascade="all,delete-orphan"),
                                   primaryjoin=(_zone_id_column == foreign(_stop_zone_id_column)) & (_zone_feed_id_column == _stop_feed_id_column))
        }))

        _transfer_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _transfer_from_stop_id_column = Column('from_stop_id', String, primary_key=True)
        _transfer_to_stop_id_column = Column('to_stop_id', String, primary_key=True)
        _transfer_mapper = Table('transfers', self._metadata,
                    _transfer_feed_id_column,
                    _transfer_from_stop_id_column,
                    _transfer_to_stop_id_column,
                    Column('transfer_type', Integer, nullable=False),
                    Column('min_transfer_time', Integer),
                    ForeignKeyConstraint(['feed_id', 'from_stop_id'], ['stops.feed_id', 'stops.stop_id']),
                    ForeignKeyConstraint(['feed_id', 'to_stop_id'], ['stops.feed_id', 'stops.stop_id']),
                    Index('idx_transfer_from', 'feed_id', 'from_stop_id'),
                    Index('idx_transfer_to', 'feed_id', 'to_stop_id'))
        self.mappers.append(mapper(Transfer, _transfer_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('transfers', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_transfer_feed_id_column)),
            'from_stop' : relationship(Stop, backref=backref('from_transfers', cascade='all', uselist=True), uselist=False,
                                       primaryjoin=(_transfer_from_stop_id_column == foreign(_stop_id_column)) & (_transfer_feed_id_column == _stop_feed_id_column)),
            'to_stop' : relationship(Stop, backref=backref('to_transfers', cascade='all', uselist=True), uselist=False,
                                     primaryjoin=(_transfer_to_stop_id_column == foreign(_stop_id_column)) & (_transfer_feed_id_column == _stop_feed_id_column))
        }))

        _route_id_column = Column('route_id', String, primary_key=True)
        _route_mapper = Table('routes', self._metadata,
                    _route_feed_id_column,
                    _route_id_column,
                    _route_agency_id_column,
                    Column('route_short_name', String),
                    Column('route_long_name', String),
                    Column('route_desc', String),
                    Column('route_type', Integer, nullable=False),
                    Column('route_url', String),
                    Column('route_color', String),
                    Column('route_text_color', String),
                    ForeignKeyConstraint(['feed_id', 'agency_id'], ['agency.feed_id', 'agency.agency_id']),
                    Index('idx_routes_agency', 'feed_id', 'agency_id'),
                    Index('idx_routes_short_name', 'feed_id', 'route_short_name'),
                    Index('idx_routes_type', 'feed_id', 'route_type'))
        self.mappers.append(mapper(Route, _route_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('routes', cascade="all,delete-orphan"),
                                    primaryjoin=_feedinfo_id_column == foreign(_route_feed_id_column)),
            'agency' : relationship(Agency, backref=backref('routes', cascade="all,delete-orphan"),
                                    primaryjoin=(_agency_id_column == foreign(_route_agency_id_column)) & (_agency_feed_id_column == _route_feed_id_column))
        }))

        _calendar_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _calendar_id_column = Column('service_id', String, primary_key=True)
        _calendar_mapper = Table('calendar', self._metadata,
                    _calendar_feed_id_column,
                    _calendar_id_column
                    )
        self.mappers.append(mapper(Calendar, _calendar_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('calendars', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_calendar_feed_id_column))
        }))

        _calendar_date_mapper = Table('calendar_dates', self._metadata,
                    Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True),
                    Column('service_id', String, primary_key=True),
                    Column('date', Date, primary_key=True),
                    ForeignKeyConstraint(['feed_id', 'service_id'], ['calendar.feed_id', 'calendar.service_id']),
                    # TOCHECK It seems a composite primary key on (a,b,c) does not need indexing on left elements,
                    # such as (a) and (a,b); but need on (a,c) for example.
                    Index('idx_calendar_dates_date', 'feed_id', 'date'))
        self.mappers.append(mapper(CalendarDate, _calendar_date_mapper, properties={
            'calendar' : relationship(Calendar, backref=backref('dates', cascade="all,delete-orphan"))
        }))

        _shape_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _shape_id_column = Column('shape_id', String, primary_key=True)
        _shape_mapper = Table('shapes', self._metadata,
                    _shape_feed_id_column,
                    _shape_id_column
                    )
        self.mappers.append(mapper(Shape, _shape_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('shapes', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_shape_feed_id_column))
        }))

        _shape_pt_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _shape_pt_shape_id_column = Column('shape_id', String, primary_key=True)
        _shape_pt_seq_column = Column('shape_pt_sequence', Integer, primary_key=True)
        _shape_pt_mapper = Table('shape_pts', self._metadata,
                    _shape_pt_feed_id_column,
                    _shape_pt_shape_id_column,
                    _shape_pt_seq_column,
                    Column('shape_dist_traveled', Float, nullable=False),
                    Column('shape_pt_lat', Float, nullable=False),
                    Column('shape_pt_lon', Float, nullable=False),
                    ForeignKeyConstraint(['feed_id', 'shape_id'], ['shapes.feed_id', 'shapes.shape_id']),
                    Index('idx_shape_pt_shape', 'feed_id', 'shape_id'))
        self.mappers.append(mapper(ShapePoint, _shape_pt_mapper, properties={
            # Note: here we specify foreign() on shape_pt feed_id column as there is no ownership relation of feed to shape_pts
            'shape' : relationship(Shape, backref=backref('points', order_by=_shape_pt_seq_column, cascade="all,delete-orphan"),
                                  primaryjoin=(_shape_id_column == foreign(_shape_pt_shape_id_column)) & (_shape_feed_id_column == foreign(_shape_pt_feed_id_column)))
        }))

        _trip_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _trip_id_column = Column('trip_id', String, primary_key=True)
        _trip_route_id_column = Column('route_id', String, nullable=False)
        _trip_calendar_id_column = Column('service_id', String, nullable=False)
        _trip_shape_id_column = Column('shape_id', String, nullable=True)
        _trip_mapper = Table('trips', self._metadata,
                    _trip_feed_id_column,
                    _trip_id_column,
                    _trip_route_id_column,
                    _trip_calendar_id_column,
                    _trip_shape_id_column,
                    Column('wheelchair_accessible', Integer, nullable=False),
                    Column('bikes_allowed', Integer, nullable=False),
                    Column('exact_times', Integer, nullable=False),
                    Column('frequency_generated', Boolean, nullable=False),
                    Column('trip_headsign', String),
                    Column('trip_short_name', String),
                    Column('direction_id', Integer),
                    Column('block_id', String),
                    ForeignKeyConstraint(['feed_id', 'route_id'], ['routes.feed_id', 'routes.route_id']),
                    ForeignKeyConstraint(['feed_id', 'service_id'], ['calendar.feed_id', 'calendar.service_id']),
                    ForeignKeyConstraint(['feed_id', 'shape_id'], ['shapes.feed_id', 'shapes.shape_id']),
                    Index('idx_trips_route', 'feed_id', 'route_id'),
                    Index('idx_trips_service', 'feed_id', 'service_id'))
        self.mappers.append(mapper(Trip, _trip_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('trips', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_trip_feed_id_column)),
            'route' : relationship(Route, backref=backref('trips', cascade="all,delete-orphan"),
                                   primaryjoin=(_route_id_column == foreign(_trip_route_id_column)) & (_route_feed_id_column == _trip_feed_id_column)),
            'calendar' : relationship(Calendar, backref=backref('trips', cascade="all,delete-orphan"),
                                      primaryjoin=(_calendar_id_column == foreign(_trip_calendar_id_column)) & (_calendar_feed_id_column == _trip_feed_id_column)),
            'shape' : relationship(Shape, backref=backref('trips', cascade="all,delete-orphan"),
                                      primaryjoin=(_shape_id_column == foreign(_trip_shape_id_column)) & (_shape_feed_id_column == _trip_feed_id_column))
        }))

        _stop_times_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _stop_times_trip_id_column = Column('trip_id', String, primary_key=True)
        _stop_seq_column = Column('stop_sequence', Integer, primary_key=True)
        _stop_times_stop_id_column = Column('stop_id', String, nullable=False)
        _stop_times_mapper = Table('stop_times', self._metadata,
                    _stop_times_feed_id_column,
                    _stop_times_trip_id_column,
                    _stop_seq_column,
                    _stop_times_stop_id_column,
                    Column('arrival_time', Integer, nullable=True),
                    Column('departure_time', Integer, nullable=True),
                    Column('interpolated', Boolean, nullable=False),
                    Column('shape_dist_traveled', Float, nullable=False),
                    Column('timepoint', Integer, nullable=False),
                    Column('pickup_type', Integer, nullable=False),
                    Column('drop_off_type', Integer, nullable=False),
                    Column('stop_headsign', String),
                    ForeignKeyConstraint(['feed_id', 'trip_id'], ['trips.feed_id', 'trips.trip_id']),
                    ForeignKeyConstraint(['feed_id', 'stop_id'], ['stops.feed_id', 'stops.stop_id']),
                    Index('idx_stop_times_stop', 'feed_id', 'stop_id'),
                    Index('idx_stop_times_sequence', 'feed_id', 'stop_sequence'))
        self.mappers.append(mapper(StopTime, _stop_times_mapper, properties={
            # Note: here we specify foreign() on stop_times feed_id column as there is no ownership relation of feed to stop_times
            'trip' : relationship(Trip, backref=backref('stop_times', order_by=_stop_seq_column, cascade="all,delete-orphan"),
                                  primaryjoin=(_trip_id_column == foreign(_stop_times_trip_id_column)) & (_trip_feed_id_column == foreign(_stop_times_feed_id_column))),
            'stop' : relationship(Stop, backref=backref('stop_times', cascade="all,delete-orphan"),
                                  primaryjoin=(_stop_id_column == foreign(_stop_times_stop_id_column)) & (_stop_feed_id_column == _stop_times_feed_id_column)),
        }))

        _fareattr_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'), primary_key=True)
        _fareattr_id_column = Column('fare_id', String, primary_key=True)
        _fareattr_mapper = Table('fare_attributes', self._metadata,
                    _fareattr_feed_id_column,
                    _fareattr_id_column,
                    Column('price', Float, nullable=False),
                    Column('currency_type', String, nullable=False),
                    Column('payment_method', Integer, nullable=False),
                    Column('transfers', Integer),
                    Column('transfer_duration', Integer))
        self.mappers.append(mapper(FareAttribute, _fareattr_mapper, properties={
            'feed' : relationship(FeedInfo, backref=backref('fare_attributes', cascade="all,delete-orphan"),
                                  primaryjoin=_feedinfo_id_column == foreign(_fareattr_feed_id_column))
        }))

        _farerule_feed_id_column = Column('feed_id', String, ForeignKey('feed_info.feed_id'))
        _farerule_id_column = Column('fare_id', String)
        # Use a dummy autoincrement numerical field for primary key,
        # as a primary key is mandatory, and the natural primary key
        # for the model (feed_id, fare_id, route+zones ids) do have
        # some fields that can be null.
        _farerule_rule_id_column = Column('fare_rule_id', Integer, primary_key=True, autoincrement=True)
        _farerule_route_id_column = Column('route_id', String, nullable=True)
        _farerule_origin_id_column = Column('origin_id', String, nullable=True)
        _farerule_destination_id_column = Column('destination_id', String, nullable=True)
        _farerule_contains_id_column = Column('contains_id', String, nullable=True)
        _farerule_mapper = Table('fare_rules', self._metadata,
                    _farerule_feed_id_column,
                    _farerule_id_column,
                    _farerule_rule_id_column,
                    _farerule_route_id_column,
                    _farerule_origin_id_column,
                    _farerule_destination_id_column,
                    _farerule_contains_id_column,
                    ForeignKeyConstraint(['feed_id', 'fare_id'], ['fare_attributes.feed_id', 'fare_attributes.fare_id']))
        self.mappers.append(mapper(FareRule, _farerule_mapper, properties={
            'fare_attribute' : relationship(FareAttribute, backref=backref('fare_rules', cascade="all,delete-orphan")),
            'route' : relationship(Route, backref=backref('fare_rules', cascade="all,delete-orphan"),
                            primaryjoin=(_route_id_column == foreign(_farerule_route_id_column)) & (_route_feed_id_column == _farerule_feed_id_column)),
            'origin' : relationship(Zone, backref=backref('origin_fare_rules', cascade="all,delete-orphan"),
                            primaryjoin=(_zone_id_column == foreign(_farerule_origin_id_column)) & (_zone_feed_id_column == _farerule_feed_id_column)),
            'destination' : relationship(Zone, backref=backref('destination_fare_rules', cascade="all,delete-orphan"),
                            primaryjoin=(_zone_id_column == foreign(_farerule_destination_id_column)) & (_zone_feed_id_column == _farerule_feed_id_column)),
            'contains' : relationship(Zone, backref=backref('contains_fare_rules', cascade="all,delete-orphan"),
                            primaryjoin=(_zone_id_column == foreign(_farerule_contains_id_column)) & (_zone_feed_id_column == _farerule_feed_id_column))
        }))

        self._metadata.create_all(engine)
        self._class_for_table = {}
        self._table_for_class = {}
        for _mapper in self.mappers:
            self._class_for_table[_mapper.mapped_table.name] = _mapper.class_
            self._table_for_class[_mapper.class_] = _mapper.mapped_table.name

    def class_for_table(self, tablename):
        """Return the class associated to a given table name.
           We implement ourselves this method as there does not seem a reliable way
           of getting this information from SqlAlchemy itself w/o some brittle hacking."""
        return self._class_for_table.get(tablename)

    def table_for_class(self, clazz):
        """Return the table name associated to a give entity class."""
        return self._table_for_class.get(clazz)



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
from gtfslib.spatial import orthodromic_distance, orthodromic_seg_distance,\
    SpatialClusterizer
import math
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unittest

class SimplePoint(object):
    def __init__(self, lat, lon):
        self._lat = lat
        self._lon = lon
    def lat(self):
        return self._lat
    def lon(self):
        return self._lon
    def __repr__(self):
        return "<%s(%.6f,%.6f)>" % (self.__class__.__name__, self._lat, self._lon)

class TestSpatial(unittest.TestCase):

    _NAUTICAL_MILE = 1853.248

    def setUp(self):
        pass

    def test_distance(self):
        a = SimplePoint(0, 0)
        b = SimplePoint(1, 0)
        c = SimplePoint(0, 1)
        self.assertAlmostEqual(orthodromic_distance(a, a), 0.0, 3)
        dab = orthodromic_distance(a, b)
        dac = orthodromic_distance(a, c)
        dbc = orthodromic_distance(b, c)
        self.assertAlmostEqual(dab, dac, 3)
        # This is the definition of the nautical mile
        self.assertAlmostEqual(dab / 60, self._NAUTICAL_MILE, 2)
        # Spherical triangular inequality
        self.assertTrue(dab * dab + dac * dac > dbc * dbc)
        d = SimplePoint(90, 0)
        e = SimplePoint(0, 90)
        dad = orthodromic_distance(a, d)
        dae = orthodromic_distance(a, e)
        dde = orthodromic_distance(d, e)
        self.assertAlmostEqual(dad, dae, 3)
        self.assertAlmostEqual(dae, dde, 3)
        self.assertAlmostEqual(dad, dde, 3)
        f = SimplePoint(45, 0)
        daf = orthodromic_distance(a, f)
        self.assertAlmostEqual(daf * 2, dad, 3)
        g = SimplePoint(45.0001, 0)
        h = SimplePoint(45, 0.0001)
        dfg = orthodromic_distance(f, g)
        dfh = orthodromic_distance(f, h)
        dgh = orthodromic_distance(g, h)
        self.assertAlmostEqual(dfg * math.cos(math.radians(45)), dfh, 3)
        # Not perfectly equals, but for small distance should hold
        self.assertAlmostEqual(dfg * dfg + dfh * dfh, dgh * dgh, 2)

    def test_seg_distance(self):
        a = SimplePoint(0, 0)
        daaa, daaa2 = orthodromic_seg_distance(a, a, a)
        self.assertAlmostEqual(daaa, 0.0, 3)
        self.assertAlmostEqual(daaa2, 0.0, 3)
        b = SimplePoint(1, 0)
        daab, daab2 = orthodromic_seg_distance(a, a, b)
        self.assertAlmostEqual(daab, 0.0, 3)
        self.assertAlmostEqual(daab2, 0.0, 3)
        dbab, dbab2 = orthodromic_seg_distance(b, a, b)
        self.assertAlmostEqual(dbab, 0.0, 3)
        self.assertAlmostEqual(dbab2 / 60, self._NAUTICAL_MILE, 2)
        c = SimplePoint(0.5, 0)
        dcab, dcab2 = orthodromic_seg_distance(c, a, b)
        self.assertAlmostEqual(dcab, 0.0, 3)
        self.assertAlmostEqual(dcab2 / 60, self._NAUTICAL_MILE / 2.0, 3)
        d = SimplePoint(-1, 0)
        ddab, ddab2 = orthodromic_seg_distance(d, a, b)
        self.assertAlmostEqual(ddab / 60, self._NAUTICAL_MILE, 2)
        print(ddab2)
        self.assertAlmostEqual(ddab2, 0, 2)
        e = SimplePoint(2, 0)
        deab, deab2 = orthodromic_seg_distance(e, a, b)
        self.assertAlmostEqual(deab / 60, self._NAUTICAL_MILE, 2)
        self.assertAlmostEqual(deab2 / 60, self._NAUTICAL_MILE, 2)
        f = SimplePoint(0.01, 1)
        dfab, dfab2 = orthodromic_seg_distance(f, a, b)
        self.assertAlmostEqual(dfab / 60, self._NAUTICAL_MILE, 2)
        self.assertAlmostEqual(dfab2 / 60, self._NAUTICAL_MILE * 0.01, 2)
        g = SimplePoint(1, 1)
        h = SimplePoint(0.5, 0.5)
        dhag, dhag2 = orthodromic_seg_distance(h, a, g)
        self.assertAlmostEqual(dhag, 0, 3)
        self.assertAlmostEqual(dhag2 / 60, self._NAUTICAL_MILE / 2 * math.sqrt(2), 0)
        # Please note that the following is true only because
        # the distance is an approximation on the equirectangular projection.
        dbag, dbag2 = orthodromic_seg_distance(b, a, g)
        self.assertAlmostEqual(dbag / 60 * math.sqrt(2), self._NAUTICAL_MILE, 0)
        self.assertAlmostEqual(dbag2 / 60, self._NAUTICAL_MILE / 2 * math.sqrt(2), 0)

    def test_clusterizer(self):

        p1 = SimplePoint(45, 0)
        p2 = SimplePoint(45 + 1.001/60, 0)
        p3 = SimplePoint(45 - 0.999/60, 0)
        sc = SpatialClusterizer(self._NAUTICAL_MILE)
        sc.add_points((p1, p2, p3))
        sc.clusterize()
        self.assertFalse(sc.in_same_cluster(p1, p2))
        self.assertTrue(sc.in_same_cluster(p1, p3))
        self.assertFalse(sc.in_same_cluster(p2, p3))
        self.assertTrue(len(sc.clusters()) == 2)

        p1 = SimplePoint(45, 0)
        p2 = SimplePoint(45 + 2*0.8/60, 0)
        p3 = SimplePoint(45 + 1*0.8/60, 0)
        sc = SpatialClusterizer(self._NAUTICAL_MILE)
        sc.add_points((p1, p2, p3))
        sc.clusterize()
        self.assertTrue(sc.in_same_cluster(p1, p2))
        self.assertTrue(sc.in_same_cluster(p1, p3))
        self.assertTrue(sc.in_same_cluster(p2, p3))
        self.assertTrue(len(sc.clusters()) == 1)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
import six
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import datetime
from functools import total_ordering


class FeedInfo(object):
    
    def __init__(self, feed_id, **kwargs):
        self.feed_id = feed_id
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __repr__(self):
        return "<%s(id=%s)>" % (
                self.__class__.__name__, self.feed_id)

class Agency(object):
    
    def __init__(self, feed_id, agency_id, agency_name, agency_url, agency_timezone, **kwargs):
        self.feed_id = feed_id
        self.agency_id = agency_id
        self.agency_name = agency_name
        self.agency_url = agency_url
        self.agency_timezone = agency_timezone
        for key in kwargs:
            setattr(self, key, kwargs[key])
        
    def __repr__(self):
        return "<%s(id=%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.agency_id, _public_vars(self))

class Zone(object):

    def __init__(self, feed_id, zone_id):
        self.feed_id = feed_id
        self.zone_id = zone_id

    def __repr__(self):
        return "<%s(id=%s/%s)>" % (
                self.__class__.__name__, self.feed_id, self.zone_id)

class Stop(object):
    
    TYPE_STOP = 0
    TYPE_STATION = 1
    
    WHEELCHAIR_UNKNOWN = 0
    WHEELCHAIR_YES = 1
    WHEELCHAIR_NO = 2

    def __init__(self, feed_id, stop_id, stop_name, stop_lat, stop_lon,
                 location_type=TYPE_STOP,
                 parent_station_id=None,
                 wheelchair_boarding=WHEELCHAIR_UNKNOWN, **kwargs):
        self.feed_id = feed_id
        self.stop_id = stop_id
        self.location_type = location_type
        self.stop_name = stop_name
        self.stop_lat = stop_lat
        self.stop_lon = stop_lon
        self.parent_station_id = parent_station_id
        self.wheelchair_boarding = wheelchair_boarding
        for key in kwargs:
            setattr(self, key, kwargs[key])
            
    def lat(self):
        return self.stop_lat

    def lon(self):
        return self.stop_lon

    def in_same_station(self, stop2):
        if self.feed_id != stop2.feed_id:
            return False
        stationid1 = self.stop_id if self.location_type == Stop.TYPE_STATION else self.parent_station_id
        stationid2 = stop2.stop_id if stop2.location_type == Stop.TYPE_STATION else stop2.parent_station_id
        return stationid1 is not None and stationid1 == stationid2

    def __repr__(self):
        return "<%s(id=%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.stop_id, _public_vars(self))

class Transfer(object):

    TRANSFER_DEFAULT = 0
    TRANSFER_SYNC = 1
    TRANSFER_TIMED = 2
    TRANSFER_NONE = 3

    def __init__(self, feed_id, from_stop_id, to_stop_id, transfer_type=TRANSFER_DEFAULT, min_transfer_time=None):
        self.feed_id = feed_id
        self.from_stop_id = from_stop_id
        self.to_stop_id = to_stop_id
        self.transfer_type = transfer_type
        self.min_transfer_time = min_transfer_time

    def __repr__(self):
        return "<%s(id=%s/%s>%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.from_stop_id, self.to_stop_id, _public_vars(self))

class Route(object):

    TYPE_TRAM = 0
    TYPE_SUBWAY = 1
    TYPE_RAIL = 2
    TYPE_BUS = 3
    TYPE_FERRY = 4
    TYPE_CABLECAR = 5
    TYPE_GONDOLA = 6
    TYPE_FUNICULAR = 7

    def __init__(self, feed_id, route_id, agency_id, route_type, **kwargs):
        self.feed_id = feed_id
        self.route_id = route_id
        self.agency_id = agency_id
        self.route_type = route_type
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def name(self, maxlen=None):
        """Will return a name based on short name, or long name if a short name is not present.
           Truncate the name to maxlen if a size is given."""
        if self.route_short_name is not None:
            ret = self.route_short_name
        else:
            ret = self.route_long_name
        if maxlen:
            ret = ret[:maxlen]
        return ret

    def __repr__(self):
        return "<%s(id=%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.route_id, _public_vars(self))

class Calendar(object):
    
    def __init__(self, feed_id, service_id):
        self.feed_id = feed_id
        self.service_id = service_id
        
    def __repr__(self):
        return "<%s(id=%s/%s)>" % (
                self.__class__.__name__, self.feed_id, self.service_id)

@total_ordering
class CalendarDate(object):

    def __init__(self, date):
        self.feed_id = None
        self.date = date
        
    @classmethod
    def ymd(cls, year, month, day):
        return cls(datetime.date(year, month, day))
    
    @classmethod
    def fromYYYYMMDD(cls, yyyymmdd):
        return cls.ymd(int(yyyymmdd[:4]), int(yyyymmdd[4:6]), int(yyyymmdd[6:8]))

    @classmethod
    def range(cls, start, end):
        cursor = start
        while cursor < end:
            yield cursor
            cursor = cursor.next_day()
            
    def next_day(self, ndays=1):
        return CalendarDate(self.date + datetime.timedelta(days=ndays))
    
    def dow(self):
        return self.date.weekday()
    
    def _coerce(self, other):
        if isinstance(other, str):
            ymd = other.split("-")
            return datetime.date(int(ymd[0]), int(ymd[1]), int(ymd[2]))
        if isinstance(other, datetime.date):
            return other
        if isinstance(other, CalendarDate):
            return other.date
        raise ValueError("Can't coerce %s to a CalendarDate (use either a CalendarDate, a 'yyyy-mm-dd' string or a datetime.date)" % other)
    
    def as_date(self):
        # Note: add a getter if internal representation change some day
        return self.date

    def toYYYYMMDD(self):
        """Return date as GTFS date format (YYYYMMDD)"""
        return "%04d%02d%02d" % (self.date.year, self.date.month, self.date.day)

    def __lt__(self, other):
        return self.date < self._coerce(other)

    def __eq__(self, other):
        return self.date == self._coerce(other)

    def __ne__(self, other):
        return self.date != self._coerce(other)
    
    def __hash__(self):
        return self.date.year * 384 + self.date.month * 32 + self.date.day
    
    def __repr__(self):
        if hasattr(self, 'feed_id') and self.feed_id is not None and hasattr(self, 'service_id') and self.service_id is not None:
            return "<%s(id=%s/%s %s)>" % (
                self.__class__.__name__, self.feed_id, self.service_id, self.date)
        else:
            return "<%s(%s)>" % (
                self.__class__.__name__, self.date)

class Trip(object):

    # Same values as Stop, but duplicated as they may differ one day    
    WHEELCHAIR_UNKNOWN = 0
    WHEELCHAIR_YES = 1
    WHEELCHAIR_NO = 2
    
    # Note: YES here really means "at least one"
    BIKES_UNKNOWN = 0
    BIKES_YES = 1
    BIKES_NO = 2
    
    TIME_APPROX = 0
    TIME_EXACT = 1
    
    def __init__(self, feed_id, trip_id, route_id, service_id,
                 wheelchair_accessible=WHEELCHAIR_UNKNOWN,
                 bikes_allowed=BIKES_UNKNOWN,
                 exact_times=TIME_EXACT,
                 frequency_generated=False, **kwargs):
        self.feed_id = feed_id
        self.trip_id = trip_id
        self.route_id = route_id
        self.service_id = service_id
        self.wheelchair_accessible = wheelchair_accessible
        self.bikes_allowed = bikes_allowed
        self.exact_times = exact_times
        self.frequency_generated = frequency_generated
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def hops(self):
        # Python rocks for this kind of stuff
        return zip(self.stop_times[0:], self.stop_times[1:])

    def __repr__(self):
        return "<%s(id=%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.trip_id, _public_vars(self))

@total_ordering
class StopTime(object):
    
    PICKUP_DROPOFF_REGULAR = 0
    PICKUP_DROPOFF_NONE = 1
    PICKUP_DROPOFF_PHONE = 2
    PICKUP_DROPOFF_ASKDRIVER = 3
    
    TIMEPOINT_APPROX = 0
    TIMEPOINT_EXACT = 1

    def __init__(self, feed_id, trip_id, stop_id, stop_sequence, arrival_time, departure_time,
                shape_dist_traveled, interpolated=False,
                timepoint=TIMEPOINT_EXACT,
                pickup_type=PICKUP_DROPOFF_REGULAR,
                drop_off_type=PICKUP_DROPOFF_REGULAR, **kwargs):
        self.feed_id = feed_id
        self.trip_id = trip_id
        self.stop_id = stop_id
        self.stop_sequence = stop_sequence
        self.arrival_time = arrival_time
        self.departure_time = departure_time
        self.interpolated = interpolated
        self.shape_dist_traveled = shape_dist_traveled
        self.timepoint = timepoint
        self.pickup_type = pickup_type
        self.drop_off_type = drop_off_type
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __lt__(self, other):
        return self.stop_sequence < other.stop_sequence

    def __eq__(self, other):
        if not isinstance(other, StopTime):
            return False
        return _generic_eq(self._primary_keys(), other._primary_keys())
        
    def __hash__(self):
        return _generic_hash(self._primary_keys())
    
    def _primary_keys(self):
        return (self.feed_id, self.trip_id, self.stop_id, self.stop_sequence)

    def __repr__(self):
        return "<%s(id=%s/%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.trip_id, self.stop_sequence, _public_vars(self))

class Shape(object):

    def __init__(self, feed_id, shape_id):
        self.feed_id = feed_id
        self.shape_id = shape_id

    def __repr__(self):
        return "<%s(id=%s/%s)>" % (
                self.__class__.__name__, self.feed_id, self.shape_id)

@total_ordering
class ShapePoint(object):

    def __init__(self, feed_id, shape_id, shape_pt_sequence, shape_pt_lat, shape_pt_lon,
                shape_dist_traveled):
        self.feed_id = feed_id
        self.shape_id = shape_id
        self.shape_pt_sequence = shape_pt_sequence
        self.shape_pt_lat = shape_pt_lat
        self.shape_pt_lon = shape_pt_lon
        self.shape_dist_traveled = shape_dist_traveled

    def lat(self):
        return self.shape_pt_lat

    def lon(self):
        return self.shape_pt_lon

    def __lt__(self, other):
        return self.shape_pt_sequence < other.shape_pt_sequence

    def __eq__(self, other):
        if not isinstance(other, ShapePoint):
            return False
        return _generic_eq(self._primary_keys(), other._primary_keys())

    def __hash__(self):
        return _generic_hash(self._primary_keys())

    def _primary_keys(self):
        return (self.feed_id, self.shape_id, self.shape_pt_sequence)

    def __repr__(self):
        return "<%s(id=%s/%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.shape_id, self.shape_pt_sequence, _public_vars(self))

class FareAttribute(object):

    PAYMENT_ONBOARD = 0
    PAYMENT_BEFOREBOARDING = 1

    def __init__(self, feed_id, fare_id, price, currency_type, payment_method, transfers, transfer_duration=None):
        self.feed_id = feed_id
        self.fare_id = fare_id
        self.price = price
        self.currency_type = currency_type
        self.payment_method = payment_method
        self.transfers = transfers
        self.transfer_duration = transfer_duration

    def __repr__(self):
        return "<%s(id=%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.fare_id, _public_vars(self))

class FareRule(object):

    def __init__(self, feed_id, fare_id, route_id=None, origin_id=None, destination_id=None, contains_id=None):
        self.feed_id = feed_id
        self.fare_id = fare_id
        self.route_id = route_id
        self.origin_id = origin_id
        self.destination_id = destination_id
        self.contains_id = contains_id

    def __eq__(self, other):
        if not isinstance(other, FareRule):
            return False
        return _generic_eq(self._primary_keys(), other._primary_keys())

    def __hash__(self):
        return _generic_hash(self._primary_keys())

    def _primary_keys(self):
        return (self.feed_id, self.fare_id, self.route_id, self.origin_id, self.destination_id, self.contains_id)

    def __repr__(self):
        return "<%s(id=%s/%s, %s)>" % (
                self.__class__.__name__, self.feed_id, self.fare_id, _public_vars(self))

# ---------------------------
# Some private helper methods

def _public_vars(obj):
    # By default only include primitive types
    # Otherwise there will be too much data (think "route.trips"...)
    PRIMITIVE_TYPES = (str, unicode, int, float, bool) if six.PY2 else (str, int, float, bool)
    return { k: v for k, v in vars(obj).items() if k[0] != '_' and isinstance(v, PRIMITIVE_TYPES) }

def _generic_hash(ids):
    h = 0
    for myid in ids:
        h += myid.__hash__()
        h *= 97
    return h

def _generic_eq(a_ids, b_ids):
    for a_id, b_id in zip(a_ids, b_ids):
        if a_id != b_id:
            return False
    return True




# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unittest

from gtfslib.model import CalendarDate, Stop
import datetime


class TestModel(unittest.TestCase):

    def setUp(self):
        pass

    def test_calendar_date(self):
        self.assertEqual(True, True)
        d1 = CalendarDate.ymd(2015, 12, 31)
        d2 = CalendarDate.ymd(2016, 1, 1)
        self.assertEquals(d1.next_day(), d2)
        self.assertTrue(d1 < d2)
        self.assertTrue(d1 <= d2)
        self.assertFalse(d1 > d2)
        self.assertFalse(d1 >= d2)
        self.assertFalse(d1 == d2)
        dt = datetime.date(2015, 12, 31)
        self.assertTrue(d1 == dt)
        dates = [ CalendarDate.ymd(2016, 1, 1), CalendarDate.ymd(2016, 1, 2) ]
        self.assertTrue(CalendarDate.ymd(2016, 1, 2) in dates)

    def test_calendar_date_set(self):
        d1 = CalendarDate.ymd(2015, 12, 31)
        d2 = CalendarDate.ymd(2016, 1, 1)
        dates = set([ d1, d2 ])
        self.assertTrue(len(dates) == 2)
        d2b = CalendarDate.ymd(2016, 1, 1)
        dates.add(d2b)
        self.assertTrue(len(dates) == 2)
        d1b = CalendarDate.ymd(2015, 12, 31)
        d4 = CalendarDate.ymd(2015, 1, 2)
        dates.add(d1b)
        dates.add(d4)
        self.assertTrue(len(dates) == 3)

    def test_calendar_date_convert(self):
        d1 = CalendarDate.fromYYYYMMDD("20151231")
        d2 = CalendarDate.ymd(2015, 12, 31)
        self.assertTrue(d1 == d2)

    def test_calendar_date_out_of_range(self):
        broke = False
        try:
            d1 = CalendarDate.ymd(2015, 12, 32)  # @UnusedVariable
        except(ValueError):
            broke = True
        self.assertTrue(broke)

    def test_calendar_date_range(self):
        d1 = CalendarDate.ymd(2016, 1, 1)
        d2 = CalendarDate.ymd(2016, 2, 1)
        n = 0
        for d in CalendarDate.range(d1, d2):
            self.assertTrue(d >= d1)
            self.assertTrue(d < d2)
            n += 1
        self.assertEqual(n, 31)

    def test_same_station(self):
        s1a = Stop('F1', 'Sa', 'StopA', 45, 0)
        s1b = Stop('F1', 'Sb', 'StopB', 45, 0.1)
        s1 = Stop('F1', 'S', 'Stop', 45, 0.05, location_type=Stop.TYPE_STATION)
        s1a.parent_station_id = 'S'
        s1b.parent_station_id = 'S'
        self.assertTrue(s1a.in_same_station(s1b))
        self.assertTrue(s1b.in_same_station(s1a))
        self.assertTrue(s1.in_same_station(s1a))
        self.assertTrue(s1a.in_same_station(s1))
        s1c = Stop('F2', 'Sb', 'StopB', 45, 0.1)
        s1c.parent_station_id = 'S'
        self.assertFalse(s1c.in_same_station(s1b))
        self.assertFalse(s1c.in_same_station(s1a))
        self.assertFalse(s1a.in_same_station(s1c))

if __name__ == '__main__':
    unittest.main()



"""
@author: Laurent GREGOIRE <laurent.gregoire@mecatran.com>
"""

from setuptools import setup, find_packages
import gtfslib


setup(
    name='gtfslib',
    version=gtfslib.__version__,
    description="GTFS database library",
    long_description="An open source library for reading, databasing, querying and manipulating GTFS-based transit data",
    url='https://github.com/afimb/gtfslib-python',
    author='AFIMB / CEREMA / MECATRAN',
    author_email='laurent.gregoire@mecatran.com',
    license='GPLv3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: GIS',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],
    keywords='GTFS transit',
    packages=find_packages(),
    install_requires=['sqlalchemy', 'six', 'docopt', 'requests', 'pyqtree', 'pyshp'],
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
    entry_points={
        'console_scripts': [
            'gtfsdbloader=gtfslib.gtfsdbloader:main',
            'gtfsrun=gtfsplugins.gtfsrun:main'
        ],
    },
)



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unittest

from gtfslib.spatial import RectangularArea
from sqlalchemy.orm import clear_mappers

from gtfslib.dao import Dao
from gtfslib.model import CalendarDate, FeedInfo, Agency, Route, Calendar, Stop, \
    Trip, StopTime, Transfer, Shape, ShapePoint, Zone, FareAttribute, FareRule

class TestDao(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        clear_mappers()

    def test_entities_creation(self):
        dao = Dao()
        f1 = FeedInfo("F1")
        a1 = Agency("F1", "A1", "Agency 1", agency_url="http://www.agency.fr/", agency_timezone="Europe/Paris")
        r1 = Route("F1", "R1", "A1", 3, route_short_name="R1", route_long_name="Route 1")
        r2 = Route("F1", "R2", "A1", 3, route_short_name="R2")
        c1 = Calendar("F1", "C1")
        c1.dates = [ CalendarDate.ymd(2015, 11, 13), CalendarDate.ymd(2015, 11, 14) ]
        dao.add_all([ f1, a1, r1, r2, c1 ])

        self.assertTrue(len(dao.feeds()) == 1)
        self.assertTrue(len(dao.agencies()) == 1)
        a1b = dao.agency("A1", feed_id="F1", prefetch_routes=True)
        self.assertTrue(a1b.agency_name == "Agency 1")
        self.assertTrue(len(a1b.routes) == 2)
        r1b = dao.route("R1", feed_id="F1")
        self.assertTrue(r1b.route_short_name == "R1")
        self.assertTrue(r1b.route_long_name == "Route 1")
        self.assertTrue(r1b.route_type == 3)
        r42 = dao.route("R42", feed_id="F1")
        self.assertTrue(r42 is None)
        
    def test_entities_deletion(self):
        dao = Dao()
        f1 = FeedInfo("F1")
        a1 = Agency("F1", "A1", "Agency 1", agency_url="http://www.agency.fr/", agency_timezone="Europe/Paris")
        dao.add_all([ f1, a1 ])
        for feed in dao.feeds():
            dao.delete(feed)
        self.assertTrue(len(dao.feeds()) == 0)

    def test_stop_station_multi_feed(self):
        dao = Dao()
        fa = FeedInfo("FA")
        fb = FeedInfo("FB")
        sa = Stop("FA", "S", "Station A", 45.0, 0.0, location_type=Stop.TYPE_STATION)
        sa1 = Stop("FA", "S1", "Stop A1", 45.0, 0.0, parent_station_id="S")
        sa2 = Stop("FA", "S2", "Stop A2", 45.0, 0.1, parent_station_id="S")
        sa3 = Stop("FA", "S3", "Stop A3", 45.0, 0.2)
        sb = Stop("FB", "S", "Station B", 45.0, 0.0, location_type=Stop.TYPE_STATION)
        sb1 = Stop("FB", "S1", "Stop B1", 45.0, 0.0, parent_station_id="S")
        sb2 = Stop("FB", "S2", "Stop B2", 45.0, 0.1, parent_station_id="S")
        dao.add_all([ fa, fb, sa, sa1, sa2, sa3, sb1, sb2, sb ])
        
        sa = dao.stop("S", feed_id="FA")
        self.assertTrue(sa.stop_name == "Station A")
        self.assertTrue(len(sa.sub_stops) == 2)
        for ssa in sa.sub_stops:
            self.assertTrue(ssa.stop_name.startswith("Stop A"))
            self.assertTrue(ssa.parent_station.stop_name == "Station A")
        
        sa1 = dao.stop("S1", feed_id="FA")
        self.assertTrue(sa1.stop_name == "Stop A1")
        self.assertTrue(sa1.parent_station.stop_name == "Station A")
        
        self.assertTrue(len(list(dao.stops())) == 7)

    def test_route_agency_multi_feed(self):
        dao = Dao()
        fa = FeedInfo("FA")
        aa1 = Agency("FA", "A", "Agency A", agency_url="http://www.agency.fr/", agency_timezone="Europe/Paris")
        ar1 = Route("FA", "R", "A", 3, route_short_name="RA", route_long_name="Route A")
        ar2 = Route("FA", "R2", "A", 3, route_short_name="RA2", route_long_name="Route A2")
        fb = FeedInfo("FB")
        ba1 = Agency("FB", "A", "Agency B", agency_url="http://www.agency.fr/", agency_timezone="Europe/Paris")
        br1 = Route("FB", "R", "A", 3, route_short_name="RB", route_long_name="Route B")
        dao.add_all([ fa, aa1, ar1, ar2, fb, ba1, br1 ])
        
        fa = dao.feed("FA")
        self.assertTrue(len(fa.agencies) == 1)
        for a in fa.agencies:
            self.assertTrue(a.agency_name == "Agency A")
        self.assertTrue(len(fa.routes) == 2)
        for r in fa.routes:
            self.assertTrue(r.route_short_name.startswith("RA"))
            self.assertTrue(r.agency.agency_name == "Agency A")

    def test_trip(self):
        dao = Dao()
        f1 = FeedInfo("F1")
        a1 = Agency("F1", "A1", "Agency 1", agency_url="http://www.agency.fr/", agency_timezone="Europe/Paris")
        r1 = Route("F1", "R1", "A1", 3, route_short_name="R1", route_long_name="Route 1")
        c1 = Calendar("F1", "C1")
        c1.dates = [ d for d in CalendarDate.range(CalendarDate.ymd(2016, 1, 1), CalendarDate.ymd(2016, 1, 31).next_day()) ]
        s1 = Stop("F1", "S1", "Stop 1", 45.0, 0.0)
        s2 = Stop("F1", "S2", "Stop 2", 45.1, 0.1)
        s3 = Stop("F1", "S3", "Stop 3", 45.2, 0.2)
        t1 = Trip("F1", "T1", "R1", "C1")
        t1.direction_id = 0
        t11 = StopTime("F1", "T1", "S1", 0, 28800, 28800, 0.0)
        t12 = StopTime("F1", "T1", "S2", 1, 29400, 29400, 0.0)
        t13 = StopTime("F1", "T1", "S3", 2, 30000, 30000, 0.0)
        t2 = Trip("F1", "T2", "R1", "C1")
        t2.direction_id = 1
        # Order is not important for now
        t2.stop_times.append(StopTime(None, None, "S1", 1, 31000, 31000, 0.0))
        t2.stop_times.append(StopTime(None, None, "S2", 0, 30600, 30600, 0.0))

        dao.add_all([ f1, a1, r1, c1, s1, s2, s3, t1, t11, t12, t13, t2 ])
        # Commit is needed to re-order stop times of T2
        dao.commit()

        cal = dao.calendar("C1", feed_id="F1")
        for trip in cal.trips:
            self.assertTrue(trip.calendar.service_id == "C1")
            for stoptime in trip.stop_times:
                self.assertTrue(stoptime.trip.calendar.service_id == "C1")
            
        stop = dao.stop("S2", feed_id="F1")
        for stoptime in stop.stop_times:
            self.assertTrue(stoptime.stop.stop_id == "S2")
            self.assertTrue(stoptime.trip.trip_id.startswith("T"))
            
        trip = dao.trip("T1", feed_id="F1")
        self.assertTrue(len(trip.stop_times) == 3)

        trip = dao.trip("T2", feed_id="F1")
        self.assertTrue(len(trip.stop_times) == 2)

        for trip in dao.trips(prefetch_stop_times=True):
            last_stop_seq = -1
            for stoptime in trip.stop_times:
                self.assertTrue(stoptime.stop_sequence > last_stop_seq)
                last_stop_seq = stoptime.stop_sequence

        for trip in dao.trips():
            for stoptime1, stoptime2 in trip.hops():
                self.assertTrue(stoptime1.trip == stoptime2.trip)
                self.assertTrue(stoptime1.stop_sequence + 1 == stoptime2.stop_sequence)

        trips = list(dao.trips(fltr=Trip.direction_id == 0))
        self.assertTrue(len(trips) == 1)
        trips = list(dao.trips(fltr=Trip.direction_id == 1))
        self.assertTrue(len(trips) == 1)

    def test_stop_station(self):
        dao = Dao()
        f1 = FeedInfo("F1")
        sa = Stop("F1", "SA", "Station A", 45.0000, 0.0000)
        sa1 = Stop("F1", "SA1", "Stop A1", 45.0001, 0.0001)
        sa1.parent_station_id = 'SA'
        sb = Stop("F1", "SB", "Station B", 45.0002, 0.0002)
        sb1 = Stop("F1", "SB1", "Stop B1", 45.0003, 0.0003)
        sb1.parent_station_id = 'SB'
        sb2 = Stop("F1", "SB2", "Stop B2", 45.0002, 0.0003)
        sb2.parent_station_id = 'SB'
        a1 = Agency("F1", "A1", "Agency 1", "url1", "Europe/Paris")
        r1 = Route("F1", "R1", "A1", Route.TYPE_BUS)
        c1 = Calendar("F1", "C1")
        t1 = Trip("F1", "T1", "R1", "C1")
        st1a = StopTime("F1", "T1", "SA1", 0, None, 3600, 0.0)
        st1b = StopTime("F1", "T1", "SB1", 1, 3800, None, 100.0)
        dao.add_all([ f1, sa, sa1, sb, sb1, sb2, a1, r1, c1, t1, st1a, st1b ])

        stops = list(dao.stops(fltr=(Agency.agency_id=='A1')))
        self.assertTrue(len(stops) == 2)
        self.assertTrue(sa1 in stops)
        self.assertTrue(sb1 in stops)

        stops = list(dao.stops(fltr=(Stop.parent_station_id=='SA')))
        self.assertTrue(len(stops) == 1)
        self.assertTrue(sa1 in stops)

        stops = list(dao.stops(fltr=(Stop.parent_station_id=='SB')))
        self.assertTrue(len(stops) == 2)
        self.assertTrue(sb1 in stops)
        self.assertTrue(sb2 in stops)

    def test_transfers(self):
        dao = Dao()
        f1 = FeedInfo("F1")
        s1 = Stop("F1", "S1", "Stop 1", 45.0000, 0.0000)
        s2 = Stop("F1", "S2", "Stop 2", 45.0001, 0.0001)
        s3 = Stop("F1", "S3", "Stop 3", 45.0002, 0.0002)
        t12 = Transfer("F1", "S1", "S2")
        t21 = Transfer("F1", "S2", "S1")
        t23 = Transfer("F1", "S2", "S3", transfer_type=Transfer.TRANSFER_TIMED, min_transfer_time=180)
        t32 = Transfer("F1", "S3", "S2", transfer_type=Transfer.TRANSFER_TIMED, min_transfer_time=120)
        t13 = Transfer("F1", "S1", "S3", transfer_type=Transfer.TRANSFER_NONE)
        a1 = Agency("F1", "A1", "Agency 1", "url1", "Europe/Paris")
        a2 = Agency("F1", "A2", "Agency 2", "url2", "Europe/London")
        r1 = Route("F1", "R1", "A1", Route.TYPE_BUS)
        r2 = Route("F1", "R2", "A2", Route.TYPE_BUS)
        c1 = Calendar("F1", "C1")
        t1 = Trip("F1", "T1", "R1", "C1")
        t2 = Trip("F1", "T2", "R2", "C1")
        st1a = StopTime("F1", "T1", "S1", 0, None, 3600, 0.0)
        st1b = StopTime("F1", "T1", "S2", 1, 3800, None, 100.0)
        st2a = StopTime("F1", "T2", "S1", 0, None, 4600, 0.0)
        st2b = StopTime("F1", "T2", "S3", 1, 4800, None, 100.0)
        dao.add_all([ f1, s1, s2, s3, t12, t21, t23, t32, t13, a1, a2, r1, r2, c1, t1, t2, st1a, st1b, st2a, st2b ])

        self.assertTrue(len(dao.transfers()) == 5)

        timed_transfers = dao.transfers(fltr=(Transfer.transfer_type == Transfer.TRANSFER_TIMED))
        self.assertTrue(len(timed_transfers) == 2)
        for transfer in timed_transfers:
            self.assertTrue(transfer.transfer_type == Transfer.TRANSFER_TIMED)

        s1_from_transfers = dao.transfers(fltr=(dao.transfer_from_stop().stop_name == "Stop 1"))
        self.assertTrue(len(s1_from_transfers) == 2)
        for transfer in s1_from_transfers:
            self.assertTrue(transfer.from_stop.stop_name == "Stop 1")

        s1_fromto_transfers = dao.transfers(fltr=((dao.transfer_from_stop().stop_name == "Stop 1") | (dao.transfer_to_stop().stop_name == "Stop 1")))
        self.assertTrue(len(s1_fromto_transfers) == 3)
        for transfer in s1_fromto_transfers:
            self.assertTrue(transfer.from_stop.stop_name == "Stop 1" or transfer.to_stop.stop_name == "Stop 1")

        s1 = dao.stop("S1", feed_id="F1")
        self.assertTrue(len(s1.from_transfers) == 2)
        self.assertTrue(len(s1.to_transfers) == 1)
        for transfer in s1.from_transfers:
            if transfer.to_stop.stop_id == "S2":
                self.assertTrue(transfer.transfer_type == Transfer.TRANSFER_DEFAULT)
            elif transfer.to_stop.stop_id == "S3":
                self.assertTrue(transfer.transfer_type == Transfer.TRANSFER_NONE)

        a1_stops = list(dao.stops(fltr=(Agency.agency_id=='A1')))
        self.assertTrue(len(a1_stops) == 2)
        self.assertTrue(s1 in a1_stops)
        self.assertTrue(s2 in a1_stops)

    def test_areas(self):
        dao = Dao()
        f1 = FeedInfo("F1")
        s1 = Stop("F1", "S1", "Stop 1", 45.0, 0.0)
        s2 = Stop("F1", "S2", "Stop 2", 45.1, 0.1)
        s3 = Stop("F1", "S3", "Stop 3", 45.2, 0.2)
        dao.add_all([ f1, s1, s2, s3 ])

        # Rectangular area
        stops = list(dao.stops(fltr=dao.in_area(RectangularArea(0, 0, 1, 1))))
        self.assertTrue(len(stops) == 0)
        stops = list(dao.stops(fltr=dao.in_area(RectangularArea(-90, -180, 90, 180))))
        self.assertTrue(len(stops) == 3)
        stops = list(dao.stops(fltr=dao.in_area(RectangularArea(45.05, 0.05, 45.15, 0.15))))
        self.assertTrue(len(stops) == 1)
        self.assertTrue(stops[0].stop_id == 'S2')

    def test_shapes(self):
        dao = Dao()
        f1 = FeedInfo("")
        a1 = Agency("", "A1", "Agency 1", agency_url="http://www.agency.fr/", agency_timezone="Europe/Paris")
        r1 = Route("", "R1", "A1", 3, route_short_name="R1", route_long_name="Route 1")
        c1 = Calendar("", "C1")
        c1.dates = [ CalendarDate.ymd(2016, 1, 31), CalendarDate.ymd(2016, 2, 1) ]
        s1 = Stop("", "S1", "Stop 1", 45.0, 0.0)
        s2 = Stop("", "S2", "Stop 2", 45.1, 0.1)
        s3 = Stop("", "S3", "Stop 3", 45.2, 0.2)
        t1 = Trip("", "T1", "R1", "C1")
        t1.stop_times = [ StopTime(None, None, "S1", 0, 28800, 28800, 0.0),
                          StopTime(None, None, "S2", 1, 29400, 29400, 2.0),
                          StopTime(None, None, "S3", 2, 30000, 30000, 4.0) ]
        t2 = Trip("", "T2", "R1", "C1")
        t2.stop_times = [ StopTime(None, None, "S2", 0, 30600, 30600, 0.0),
                          StopTime(None, None, "S1", 1, 31000, 31000, 1.0) ]
        sh1 = Shape("", "Sh1")
        sh1.points = [ ShapePoint(None, None, 0, 45.00, 0.00, 0.0),
                      ShapePoint(None, None, 1, 45.05, 0.10, 1.0),
                      ShapePoint(None, None, 2, 45.10, 0.10, 2.0),
                      ShapePoint(None, None, 3, 45.15, 0.20, 3.0),
                      ShapePoint(None, None, 4, 45.20, 0.20, 4.0) ]
        t1.shape = sh1
        dao.add_all([ f1, a1, r1, c1, s1, s2, s3, t1, t2, sh1 ])
        dao.commit()

        t = dao.trip("T1")
        self.assertTrue(t.shape.shape_id == "Sh1")
        self.assertTrue(len(t.shape.points) == 5)
        t = dao.trip("T2")
        self.assertTrue(t.shape == None)

    def test_zones(self):
        dao = Dao()
        f1 = FeedInfo("")
        z1 = Zone("", "Z1")
        s1 = Stop("", "S1", "Stop 1", 45.0, 0.0)
        s2 = Stop("", "S2", "Stop 2", 45.1, 0.1, zone_id="Z1")
        s3 = Stop("", "S3", "Stop 3", 45.2, 0.2)
        s3.zone = z1
        dao.add_all([ f1, z1, s1, s2, s3 ])
        dao.commit()

        self.assertTrue(len(dao.zones()) == 1)
        z = dao.zone("Z1")
        self.assertTrue(len(z.stops) == 2)
        for stop in z.stops:
            self.assertTrue(stop.zone == z)
        s = dao.stop("S1")
        self.assertTrue(s.zone == None)
        s = dao.stop("S2")
        self.assertTrue(s.zone == z)
        s = dao.stop("S3")
        self.assertTrue(s.zone == z)

    def test_fares(self):
        dao = Dao()
        f1 = FeedInfo("")
        a1 = Agency("", "A1", "Agency 1", agency_url="http://www.agency.fr/", agency_timezone="Europe/Paris")
        r1 = Route("", "R1", "A1", 3, route_short_name="R1", route_long_name="Route 1")
        r2 = Route("", "R2", "A1", 3, route_short_name="R2", route_long_name="Route 2")
        z1 = Zone("", "Z1")
        z2 = Zone("", "Z2")
        fare1 = FareAttribute("", "F1", 1.0, "EUR", FareAttribute.PAYMENT_ONBOARD, None, None)
        fare2 = FareAttribute("", "F2", 2.0, "EUR", FareAttribute.PAYMENT_BEFOREBOARDING, 3, 3600)
        rule1 = FareRule("", "F1", route_id="R1")
        rule2 = FareRule("", "F1", origin_id="Z1", destination_id="Z2")
        dao.add_all([ f1, a1, r1, r2, z1, z2, fare1, fare2, rule1, rule2 ])
        dao.commit()

        self.assertTrue(len(dao.fare_attributes()) == 2)
        self.assertTrue(len(dao.fare_rules(fltr=(FareRule.route == r1))) == 1)
        self.assertTrue(len(dao.fare_rules(fltr=(FareRule.route == r2))) == 0)
        self.assertTrue(len(dao.fare_rules(fltr=(FareRule.origin == z1))) == 1)
        fare = dao.fare_attribute("F1")
        self.assertTrue(len(fare.fare_rules) == 2)
        fare = dao.fare_attribute("F2")
        self.assertTrue(len(fare.fare_rules) == 0)

        # Test equivalence and hash on primary keys for rule
        fr1a = FareRule("", "F1", route_id="R", origin_id="ZO", destination_id="ZD", contains_id=None)
        fr1b = FareRule("", "F1", route_id="R", origin_id="ZO", destination_id="ZD", contains_id=None)
        fr2 = FareRule("", "F1", route_id="R", origin_id="ZO", destination_id=None, contains_id="ZD")
        ruleset = set()
        ruleset.add(fr1a)
        ruleset.add(fr2)
        ruleset.add(fr1b)
        self.assertTrue(len(ruleset) == 2)
        self.assertTrue(fr1a == fr1b)
        self.assertTrue(fr1a != fr2)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unittest

from sqlalchemy.orm import clear_mappers

from gtfslib.model import Trip
from gtfslib.dao import Dao

# Location of mini.gtfs.zip.
# This unit-test is highly dependent on the CONTENT of this GTFS.
MINI_GTFS = "test/mini.gtfs.zip"

DAO_URL = ""
# To unit-test with postgresql, create a db "gtfs" with user "gtfs" and uncomment the following line:
# DAO_URL = "postgresql://gtfs@localhost/gtfs"
# Set this to true to activate SQL logging
SQL_LOG = False

class TestMiniGtfs(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        clear_mappers()

    def test_gtfs_data(self):
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        dao.load_gtfs(MINI_GTFS)

        # Check feed
        feed = dao.feed()
        self.assertTrue(feed.feed_id == "")
        self.assertTrue(feed.feed_publisher_name is None)
        self.assertTrue(feed.feed_publisher_url is None)
        self.assertTrue(feed.feed_contact_email is None)
        self.assertTrue(feed.feed_contact_url is None)
        self.assertTrue(feed.feed_start_date is None)
        self.assertTrue(feed.feed_end_date is None)
        self.assertTrue(len(dao.agencies()) == 1)
        self.assertTrue(len(dao.routes()) == 1)
        self.assertTrue(len(feed.agencies) == 1)
        self.assertTrue(len(feed.routes) == 1)

        # Check if optional route agency is set
        a = dao.agency("A")
        self.assertTrue(a is not None)
        self.assertTrue(len(a.routes) == 1)

        # Check for frequency-generated trips
        # They should all have the same delta
        trips = dao.trips(fltr=(Trip.frequency_generated == True), prefetch_stop_times=True)
        n_trips = 0
        deltas = {}
        for trip in trips:
            original_trip_id = trip.trip_id.rsplit('@', 1)[0]
            delta1 = []
            for st1, st2 in trip.hops():
                delta1.append(st2.arrival_time - st1.departure_time)
            delta2 = deltas.get(original_trip_id)
            if delta2 is not None:
                self.assertTrue(delta1 == delta2)
            else:
                deltas[original_trip_id] = delta1
            n_trips += 1
        self.assertTrue(n_trips == 8)

    def test_whitespace_stripping(self):
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        dao.load_gtfs(MINI_GTFS)

        # Check if whitespace stripping works
        a = dao.agency("A")
        self.assertTrue(a.agency_name == "Mini Agency")
        self.assertTrue(a.agency_lang == "en")

    def test_hops(self):
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        dao.load_gtfs(MINI_GTFS)

        # Get all hops
        hops = dao.hops()
        nhops = 0
        for st1, st2 in hops:
            self.assertTrue(st1.stop_sequence + 1 == st2.stop_sequence)
            self.assertTrue(st1.trip == st2.trip)
            nhops += 1
        # 2 standard trips + 2 frequency generated (8 total)
        self.assertTrue(nhops == 2 * 2 + 8 * 2)

        # Get all hops with a distance <= 70km
        hops = dao.hops(fltr=(dao.hop_second().shape_dist_traveled - dao.hop_first().shape_dist_traveled <= 70000))
        nhops1 = 0
        for st1, st2 in hops:
            self.assertTrue(st1.stop_sequence + 1 == st2.stop_sequence)
            self.assertTrue(st1.trip == st2.trip)
            self.assertTrue(st2.shape_dist_traveled - st1.shape_dist_traveled <= 70000)
            nhops1 += 1

        # Get all hops with a distance > 70km
        hops = dao.hops(fltr=(dao.hop_second().shape_dist_traveled - dao.hop_first().shape_dist_traveled > 70000))
        nhops2 = 0
        for st1, st2 in hops:
            self.assertTrue(st1.stop_sequence + 1 == st2.stop_sequence)
            self.assertTrue(st1.trip == st2.trip)
            self.assertTrue(st2.shape_dist_traveled - st1.shape_dist_traveled > 70000)
            nhops2 += 1
        self.assertTrue(nhops == nhops1 + nhops2)

        # Get all hops with a time +/- 1h
        hops = dao.hops(fltr=(dao.hop_second().arrival_time - dao.hop_first().departure_time >= 3600))
        for st1, st2 in hops:
            self.assertTrue(st2.arrival_time - st1.departure_time >= 3600)
        hops = dao.hops(fltr=(dao.hop_second().arrival_time - dao.hop_first().departure_time < 3600))
        for st1, st2 in hops:
            self.assertTrue(st2.arrival_time - st1.departure_time < 3600)

        # Get hops with a delta of 2
        hops = dao.hops(delta=2)
        for st1, st2 in hops:
            self.assertTrue(st1.stop_sequence + 2 == st2.stop_sequence)
            self.assertTrue(st1.trip == st2.trip)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""
import collections
import csv
import io
import six
import zipfile

class CsvTableFactory(object):

    def __init__(self, objname, rows):
        self._header = [ self._strip_and_nullify(item) for item in six.next(rows) ]
        if self._header[0].startswith(six.u('\ufeff')):
            self._header[0] = self._header[0][1:]
        self._rows = rows
        self._factory = collections.namedtuple(objname, self._header)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self._header)

    def __iter__(self):
        return self

    def _strip_and_nullify(self, item):
        ret = item.strip()
        if len(ret) == 0:
            return None
        return ret

    def __next__(self):
        row = [ self._strip_and_nullify(item) for item in six.next(self._rows) ]
        args = dict(six.moves.zip_longest(self._header, row, fillvalue=None))
        # return self._factory(**args)
        # Directly return the dictionary, do not bother with an object creation
        # (the first thing the client does is converting the object to a dict...)
        return args
    
    def next(self):
        return self.__next__()

def python2or3_csv(filesource, filename, encoding, dialect=csv.excel, **kwargs):
    if six.PY2:
        filedata = filesource.open(filename, 'rU')
    else:
        filedata = io.TextIOWrapper(filesource.open(filename, 'r'), encoding=encoding)
    csvreader = csv.reader(filedata, dialect=dialect, **kwargs)
    for row in csvreader:
        if len(row) == 0:
            continue
        if six.PY2:
            yield [ unicode(cell, encoding) for cell in row ]
        else:
            yield row

class ZipFileSource(object):
    
    def __init__(self, inputfile):
        self._zipfile = zipfile.ZipFile(inputfile)
    
    def open(self, filename, mode='rU'):
        return self._zipfile.open(filename, mode)
    
    def close(self):
        self._zipfile.close()

class Gtfs(object):
    TABLES = [
            dict(obj='GtfsFeedInfo', getter='feedinfo', table='feed_info.txt', optional=True),
            dict(obj='GtfsAgency', getter='agencies', table='agency.txt'),
            dict(obj='GtfsStop', getter='stops', table='stops.txt'),
            dict(obj='GtfsRoute', getter='routes', table='routes.txt'),
            dict(obj='GtfsTrip', getter='trips', table='trips.txt'),
            dict(obj='GtfsStopTime', getter='stop_times', table='stop_times.txt'),
            dict(obj='GtfsCalendar', getter='calendars', table='calendar.txt', optional=True),
            dict(obj='GtfsCalendarDate', getter='calendar_dates', table='calendar_dates.txt', optional=True),
            dict(obj='GtfsTransfer', getter='transfers', table='transfers.txt', optional=True),
            dict(obj='GtfsFrequency', getter='frequencies', table='frequencies.txt', optional=True),
            dict(obj='GtfsShape', getter='shapes', table='shapes.txt', optional=True),
            dict(obj='GtfsFareAttribute', getter='fare_attributes', table='fare_attributes.txt', optional=True),
            dict(obj='GtfsFareRule', getter='fare_rules', table='fare_rules.txt', optional=True)
    ]
    
    def __init__(self, filesource):
        self._filesource = filesource
        
    def load(self):
        for tbl in Gtfs.TABLES:
            setattr(Gtfs, tbl['getter'], self.make_getter(tbl['obj'], tbl['table'], optional=tbl.get('optional')))
        return self

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self._filesource)

    def make_getter(self, name, filename, optional=False):
        def getter(self):
            try:
                myreader = python2or3_csv(self._filesource, filename, 'utf-8')
                mytable = CsvTableFactory(name, myreader)
                return mytable
            except KeyError:
                # If table is optional, return empty iterator
                if optional:
                    return iter(())
                else:
                    raise KeyError("Required table '%s' not found in GTFS." % filename)
        return getter
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self._filesource.close()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

from gtfslib.spatial import SpatialClusterizer
from gtfslib.utils import fmttime
from gtfsplugins.prettycsv import PrettyCsv
from collections import defaultdict

class Frequencies(object):
    """
    Analysis of departure frequencies for each stop (or stop cluster).
    Warning: This plugin is still in development, the interface may change.
    
    Parameters:
    --csv=<file>     Output to given file as CSV
    --cluster=<dist> Cluster stops closer than <dist> meters (default to 0)
    --dstp=<k>       Apply this penalty to clustering for stops that do not
                     share the same station (default to 0.5)
    --samename       Cluster stops only if they share the exact same name
                     (default to False).
    --alldates       To print frequencies on all filtered dates.
                     Otherwise takes date with max departures only (default).
    
    Examples:
    --filter="CalendarDate.date=='2016-01-22'"
      For all stops at a given date
    --filter="Stop.stop_name=='Villenouvelle'"
      For one stop at all dates
    --filter="(Route.route_short_name=='R1')"
      For all stops of a single route.
    """

    def __init__(self):
        pass

    def run(self, context, csv=None, cluster=0, dstp=0.5, samename=False, alldates=False, **kwargs):
        cluster_meters = float(cluster)
        dstp = float(dstp)

        print("Loading stops...")
        stops = set()
        sc = SpatialClusterizer(cluster_meters)
        for stop in context.dao().stops(fltr=context.args.filter):
            sc.add_point(stop)
            stops.add(stop)
        print("Loaded %d stops. Clusterize..." % (len(stops)))
        sc.clusterize(comparator=sc.make_comparator(samename, dstp))
        print("Aggregated in %d clusters" % (len(sc.clusters())))
        
        print("Loading calendar dates...")
        dates = set(context.dao().calendar_dates_date(fltr=context.args.filter))
        print("Loaded %d dates" % (len(dates)))
        
        print("Processing trips...")
        departures_by_clusters = defaultdict(lambda : defaultdict(list))
        ntrips = 0
        for trip in context.dao().trips(fltr=context.args.filter, prefetch_stops=True, prefetch_stop_times=True, prefetch_calendars=True):
            for stop_time in trip.stop_times:
                if not stop_time.departure_time:
                    continue
                if not stop_time.stop in stops:
                    continue
                cluster = sc.cluster_of(stop_time.stop)
                departures_by_dates = departures_by_clusters[cluster]
                for date in trip.calendar.dates:
                    if date.as_date() not in dates:
                        continue
                    departures_by_dates[date.as_date()].append(stop_time)
            if ntrips % 1000 == 0:
                print("%d trips..." % (ntrips))
            ntrips += 1

        with PrettyCsv(csv, ["cluster", "stop_id", "stop_name", "date", "departures", "min_time", "max_time", "dep_hour" ], **kwargs) as csvout:
            for cluster, departures_by_dates in departures_by_clusters.items():
                for stop in cluster.items:
                    csvout.writerow([ cluster.id, stop.stop_id, stop.stop_name ])
                if alldates:
                    # Print departure count for all dates
                    dates_to_print = list(departures_by_dates.keys())
                    dates_to_print.sort()
                else:
                    # Compute the max only
                    date_max = None
                    dep_max = 0
                    for date, departures in departures_by_dates.items():
                        ndep = len(departures)
                        if ndep >= dep_max:
                            dep_max = ndep
                            date_max = date
                    if date_max is None:
                        continue
                    dates_to_print = [ date_max ]
                for date in dates_to_print:
                    dep_times = [dep.departure_time for dep in departures_by_dates.get(date)]
                    max_hour = max(dep_times)
                    min_hour = min(dep_times)
                    delta_hour = max_hour - min_hour
                    avg_dep = float('inf') if delta_hour == 0 else len(dep_times) * 3600. / (max_hour - min_hour)
                    csvout.writerow([ cluster.id, None, None, date, len(dep_times), fmttime(min_hour), fmttime(max_hour), "%.3f" % avg_dep ])



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
from gtfslib.model import Stop, Transfer, FareAttribute, FareRule
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import os
import zipfile
import six

from gtfslib.utils import fmttime, group_pairs
from gtfsplugins.prettycsv import PrettyCsv


class GtfsExport(object):
    """
    Export some data in GTFS-compatible format.

    Parameters:
    --skip_shape_dist     To remove shape_dist_traveled from the export.
    --bundle[=<zipfile>]  Zip the result (using filename if given,
                          otherwise default to "gtfs.zip").

    Examples:
    --filter="(Route.route_short_name=='R1')"
      Restrict to route R1
    """

    def __init__(self):
        pass

    def run(self, context, skip_shape_dist=False, bundle=None, **kwargs):

        with PrettyCsv("agency.txt", ["agency_id", "agency_name", "agency_url", "agency_timezone", "agency_lang", "agency_phone", "agency_fare_url", "agency_email" ], **kwargs) as csvout:
            nagencies = 0
            for agency in context.dao().agencies(fltr=context.args.filter):
                nagencies += 1
                csvout.writerow([ agency.agency_id, agency.agency_name, agency.agency_url, agency.agency_timezone, agency.agency_lang, agency.agency_phone, agency.agency_fare_url, agency.agency_email ])
            print("Exported %d agencies" % (nagencies))

        stop_ids = set()
        zone_ids = set()
        def _output_stop(stop):
            csvout.writerow([ stop.stop_id, stop.stop_code, stop.stop_name, stop.stop_desc, stop.stop_lat, stop.stop_lon, stop.zone_id, stop.stop_url, stop.location_type, stop.parent_station_id, stop.stop_timezone, stop.wheelchair_boarding ])

        with PrettyCsv("stops.txt", ["stop_id", "stop_code", "stop_name", "stop_desc", "stop_lat", "stop_lon", "zone_id", "stop_url", "location_type", "parent_station", "stop_timezone", "wheelchair_boarding" ], **kwargs) as csvout:
            nstops = 0
            station_ids = set()
            for stop in context.dao().stops(fltr=context.args.filter, prefetch_parent=False, prefetch_substops=False):
                _output_stop(stop)
                stop_ids.add((stop.feed_id, stop.stop_id))
                if stop.parent_station_id is not None:
                    station_ids.add((stop.feed_id, stop.parent_station_id))
                if stop.zone_id is not None:
                    zone_ids.add((stop.feed_id, stop.zone_id))
                nstops += 1
            # Only export parent station that have not been already seen
            station_ids -= stop_ids
            for feed_id, st_ids in group_pairs(station_ids, 1000):
                for station in context.dao().stops(fltr=(Stop.feed_id == feed_id) & (Stop.stop_id.in_(st_ids))):
                    _output_stop(station)
                    if station.zone_id is not None:
                        zone_ids.add((station.feed_id, station.zone_id))
                    nstops += 1
            print("Exported %d stops" % (nstops))
            stop_ids |= station_ids

        route_ids = set()
        with PrettyCsv("routes.txt", ["route_id", "agency_id", "route_short_name", "route_long_name", "route_desc", "route_type", "route_url", "route_color", "route_text_color" ], **kwargs) as csvout:
            nroutes = 0
            for route in context.dao().routes(fltr=context.args.filter):
                nroutes += 1
                csvout.writerow([ route.route_id, route.agency_id, route.route_short_name, route.route_long_name, route.route_desc, route.route_type, route.route_url, route.route_color, route.route_text_color ])
                route_ids.add((route.feed_id, route.route_id))
            print("Exported %d routes" % (nroutes))

        stop_times_columns = ["trip_id", "arrival_time", "departure_time", "stop_id", "stop_sequence", "stop_headsign", "pickup_type", "drop_off_type", "timepoint"]
        if not skip_shape_dist:
            stop_times_columns.append("shape_dist_traveled")
        with PrettyCsv("trips.txt", ["route_id", "service_id", "trip_id", "trip_headsign", "trip_short_name", "direction_id", "block_id", "shape_id", "wheelchair_accessible", "bikes_allowed" ], **kwargs) as csvout1:
            with PrettyCsv("stop_times.txt", stop_times_columns, **kwargs) as csvout2:
                ntrips = 0
                nstoptimes = 0
                for trip in context.dao().trips(fltr=context.args.filter, prefetch_stops=False, prefetch_stop_times=True, prefetch_calendars=False, prefetch_routes=False):
                    ntrips += 1
                    if ntrips % 1000 == 0:
                        print("%d trips..." % (ntrips))
                    csvout1.writerow([ trip.route_id, trip.service_id, trip.trip_id, trip.trip_headsign, trip.trip_short_name, trip.direction_id, trip.block_id, trip.shape_id, trip.wheelchair_accessible, trip.bikes_allowed])
                    for stoptime in trip.stop_times:
                        nstoptimes += 1
                        row = [ trip.trip_id,
                                fmttime(stoptime.arrival_time if stoptime.arrival_time is not None else stoptime.departure_time),
                                fmttime(stoptime.departure_time if stoptime.departure_time is not None else stoptime.arrival_time),
                                stoptime.stop_id,
                                stoptime.stop_sequence,
                                stoptime.stop_headsign,
                                stoptime.pickup_type,
                                stoptime.drop_off_type,
                                stoptime.timepoint ]
                        if not skip_shape_dist:
                            row.append(stoptime.shape_dist_traveled)
                        csvout2.writerow(row)
                print("Exported %d trips with %d stop times" % (ntrips, nstoptimes))

        # Note: GTFS' model does not have calendars objects to export,
        # since a calendar is renormalized/expanded to a list of dates.

        with PrettyCsv("calendar_dates.txt", ["service_id", "date", "exception_type"], **kwargs) as csvout:
            ncals = ndates = 0
            for calendar in context.dao().calendars(fltr=context.args.filter, prefetch_dates=True):
                ncals += 1
                if ncals % 1000 == 0:
                    print("%d calendars, %d dates..." % (ncals, ndates))
                for date in calendar.dates:
                    ndates += 1
                    csvout.writerow([calendar.service_id, date.toYYYYMMDD(), 1])
            print("Exported %d calendars with %d dates" % (ncals, ndates))

        fare_attr_ids = set()
        nfarerules = [0]
        def _output_farerule(farerule):
            if farerule.route_id is not None and (farerule.feed_id, farerule.route_id) not in route_ids:
                return False
            if farerule.origin_id is not None and (farerule.feed_id, farerule.origin_id) not in zone_ids:
                return False
            if farerule.contains_id is not None and (farerule.feed_id, farerule.contains_id) not in zone_ids:
                return False
            if farerule.destination_id is not None and (farerule.feed_id, farerule.destination_id) not in zone_ids:
                return False
            csvout.writerow([ farerule.fare_id, farerule.route_id, farerule.origin_id, farerule.destination_id, farerule.contains_id ])
            fare_attr_ids.add((farerule.feed_id, farerule.fare_id))
            nfarerules[0] += 1
            return True
        with PrettyCsv("fare_rules.txt", ["fare_id", "route_id", "origin_id", "destination_id", "contains_id"], **kwargs) as csvout:
            feed_ids = set()
            for feed_id, rt_ids in group_pairs(route_ids, 1000):
                feed_ids.add(feed_id)
                for farerule in context.dao().fare_rules(fltr=(FareRule.feed_id==feed_id) & FareRule.route_id.in_(rt_ids), prefetch_fare_attributes=False):
                    if not _output_farerule(farerule):
                        continue
            for feed_id, zn_ids in group_pairs(zone_ids, 1000):
                feed_ids.add(feed_id)
                for farerule in context.dao().fare_rules(fltr=(FareRule.feed_id==feed_id) & (FareRule.origin_id.in_(zn_ids) | FareRule.contains_id.in_(zn_ids) | FareRule.destination_id.in_(zn_ids)), prefetch_fare_attributes=False):
                    if not _output_farerule(farerule):
                        continue
            # Special code to include all fare rules w/o any relationships
            # of any feed_id we've encountered so far
            for feed_id in feed_ids:
                for farerule in context.dao().fare_rules(fltr=(FareRule.feed_id==feed_id) & (FareRule.route_id==None) & (FareRule.origin_id==None) & (FareRule.contains_id==None) & (FareRule.destination_id==None), prefetch_fare_attributes=False):
                    if not _output_farerule(farerule):
                        continue
            print("Exported %d fare rules" % (nfarerules[0]))
        if nfarerules[0] == 0:
            os.remove("fare_rules.txt")

        with PrettyCsv("fare_attributes.txt", ["fare_id", "price", "currency_type", "payment_method", "transfers", "transfer_duration"], **kwargs) as csvout:
            nfareattrs = 0
            for feed_id, fa_ids in group_pairs(fare_attr_ids, 1000):
                for fareattr in context.dao().fare_attributes(fltr=(FareAttribute.feed_id==feed_id) & FareAttribute.fare_id.in_(fa_ids), prefetch_fare_rules=False):
                    nfareattrs += 1
                    csvout.writerow([ fareattr.fare_id, fareattr.price, fareattr.currency_type, fareattr.payment_method, fareattr.transfers, fareattr.transfer_duration ])
            print("Exported %d fare attributes" % (nfareattrs))
        if nfareattrs == 0:
            os.remove("fare_attributes.txt")

        shapes_columns = ["shape_id", "shape_pt_lat", "shape_pt_lon", "shape_pt_sequence"]
        if not skip_shape_dist:
            shapes_columns.append("shape_dist_traveled")
        with PrettyCsv("shapes.txt", shapes_columns, **kwargs) as csvout:
            nshapes = nshapepoints = 0
            for shape in context.dao().shapes(fltr=context.args.filter, prefetch_points=True):
                nshapes += 1
                if nshapes % 100 == 0:
                    print("%d shapes, %d points..." % (nshapes, nshapepoints))
                for point in shape.points:
                    nshapepoints += 1
                    row = [shape.shape_id, point.shape_pt_lat, point.shape_pt_lon, point.shape_pt_sequence]
                    if not skip_shape_dist:
                        row.append(point.shape_dist_traveled)
                    csvout.writerow(row)
            print("Exported %d shapes with %d points" % (nshapes, nshapepoints))
        if nshapes == 0:
            os.remove("shapes.txt")

        with PrettyCsv("transfers.txt", ["from_stop_id", "to_stop_id", "transfer_type", "min_transfer_time"], **kwargs) as csvout:
            ntransfers = 0
            transfer_ids = set()
            for feed_id, st_ids in group_pairs(stop_ids, 1000):
                # Note: we can't use a & operator below instead of |,
                # as we would need to have *all* IDs in one batch.
                for transfer in context.dao().transfers(fltr=(Transfer.feed_id == feed_id) & (Transfer.from_stop_id.in_(st_ids) | Transfer.to_stop_id.in_(st_ids)), prefetch_stops=False):
                    # As we used from_stop_id.in(...) OR to_stop_id.in(...),
                    # we need to filter out the potential superfluous results.
                    from_stop_id = (transfer.feed_id, transfer.from_stop_id)
                    to_stop_id = (transfer.feed_id, transfer.to_stop_id)
                    if from_stop_id not in stop_ids or to_stop_id not in stop_ids:
                        continue
                    transfer_id = (from_stop_id, to_stop_id)
                    if transfer_id in transfer_ids:
                        # Prevent duplicates (can happen from grouping)
                        continue
                    transfer_ids.add(transfer_id)
                    ntransfers += 1
                    csvout.writerow([ transfer.from_stop_id, transfer.to_stop_id, transfer.transfer_type, transfer.min_transfer_time ])
            print("Exported %d transfers" % (ntransfers))
        if ntransfers == 0:
            os.remove("transfers.txt")

        if bundle:
            if not isinstance(bundle, six.string_types):
                # Allow the use of "--bundle" option only
                bundle = "gtfs.zip"
            if not bundle.endswith('.zip'):
                bundle = bundle + '.zip'
            print("Zipping result to %s (removing .txt files)" % (bundle))
            with zipfile.ZipFile(bundle, 'w', zipfile.ZIP_DEFLATED) as zipf:
                for f in [ "agency.txt", "stops.txt", "routes.txt", "trips.txt", "stop_times.txt", "calendar_dates.txt", "fare_rules.txt", "fare_attributes.txt", "shapes.txt", "transfers.txt" ]:
                    if os.path.isfile(f):
                        zipf.write(f)
                        os.remove(f)